@@ -17,10 +17,62 @@ import math
 import time
 from PIL import Image
 from scipy.spatial import Delaunay
-class Face():
+
+from .Helpers import buildCameraMatrix
+
+# Get an instance of drawing specs to be used for drawing masks on faces
+DrawingSpec =  mp.solutions.drawing_utils.DrawingSpec
+class Face():    
     """Face is the class that provides operations on face landmarks.
     It is extracted by the face analyzer and could then be used for multiple face features extraction purposes
     """
+
+
+    # Key landmark indices
+    nose_tip_index = 4
+
+    left_eyelids_indices = [362, 374, 263, 386]
+    left_eye_contour_indices = [474, 475, 476, 477]
+    left_eye_center_index = 473
+
+    right_eye_contour_indices = [469, 470, 471, 472]
+    right_eyelids_indices = [130, 145, 133, 159]
+    right_eye_center_index = 468    
+
+    # Nose, left face_extremety, right_face_extremity
+    
+    # these points was chosen so that the mouth motion and eye closong do not affect them
+    # Three points were removed from my initial code (I leave them for tests) as they seem to be affected by grimacing (the chin) or are not very accurate (Left and Right)
+    face_3d_reference_positions=np.array([
+        [0,0,0],            # Nose tip
+        #[-80,50,-90],       # Left
+        #[0,-70,-30],        # Chin
+        #[80,50,-90],        # Right
+        [-70,50,-70],       # Left left eye
+        [70,50,-70],        # Right right eye
+        [0,80,-30]        # forehead center
+    ])
+    face_orientation_landmarks = [
+        4,          # Nose tip
+        #127,        # Left
+        #152,        # Chin
+        #264,        # Right
+        130,        # Left left eye
+        359,        # Right right eye
+        151         # forehead center
+        ]
+    
+
+    # A list of simplified facial features used to reduce computation cost of drawing and morphing faces
+    simplified_face_features = [
+        10, 67, 54, 162, 127, 234, 93, 132,172,150,176,148,152,377,378,365,435,323,447,454,264,389,251, 332, 338, #Oval
+        139, 105, 107, 151, 8, 9, 336, 334, 368,                            #  Eyelids
+        130, 145, 155, 6, 382, 374, 359, 159, 386,                  #  Eyes
+        129, 219, 79, 238, 2, 458, 457, 439, 358, 1, 4, 5, 197,     #  Nose
+        61, 84, 314, 409, 14, 87, 81, 12,37,267, 402, 311, 321, 269, 39, 415, 91, 178, 73, 303, 325,
+        50, 207, 280, 427
+    ]
+
     def __init__(self, landmarks:NamedTuple = None, image_shape: tuple = (480, 640)):
         """Creates an instance of Face
 
@@ -33,15 +85,6 @@ class Face():
         self.update(landmarks)
 
 
-        self.left_eyelids_indices = [362, 374, 263, 386]
-        self.left_eye_contour_indices = [474, 475, 476, 477]
-
-        self.left_eye_center_index = 473
-
-        self.right_eye_contour_indices = [469, 470, 471, 472]
-        self.right_eyelids_indices = [130, 145, 133, 159]
-        self.right_eye_center_index = 468
-
         self.blinking = False
 
         self.face_oval = list(set(
@@ -51,14 +94,7 @@ class Face():
             list(sum(list(mp.solutions.face_mesh.FACEMESH_CONTOURS), ()))[::3]
         ))
 
-        self.simplified_face_features = [
-            10, 67, 54, 162, 127, 234, 93, 132,172,150,176,148,152,377,378,365,435,323,447,454,264,389,251, 332, 338, #Oval
-            139, 105, 107, 151, 8, 9, 336, 334, 368,                            #  Eyelids
-            130, 145, 155, 6, 382, 374, 359, 159, 386,                  #  Eyes
-            129, 219, 79, 238, 2, 458, 457, 439, 358, 1, 4, 5, 197,     #  Nose
-            61, 84, 314, 409, 14, 87, 81, 12,37,267, 402, 311, 321, 269, 39, 415, 91, 178, 73, 303, 325,
-            50, 207, 280, 427
-        ]
+
 
         self.reference_facial_cloud = None
 
@@ -86,29 +122,7 @@ class Face():
             self.landmarks = None
             self.npLandmarks = np.array([])
 
-    def rotationMatrixToEulerAngles(self, R: np.ndarray) -> np.ndarray:
-        """Computes the Euler angles in the form of Pitch yaw roll
 
-        Args:
-            R (np.ndarray): The rotation matrix
-
-        Returns:
-            np.ndarray: (Pitch, Yaw, Roll)
-        """
-        sy = math.sqrt(R[0, 0] * R[0, 0] + R[1, 0] * R[1, 0])
-
-        singular = sy < 1e-6
-
-        if not singular:
-            x = math.atan2(R[2, 1], R[2, 2])
-            y = math.atan2(-R[2, 0], sy)
-            z = math.atan2(R[1, 0], R[0, 0])
-        else:
-            x = math.atan2(-R[1, 2], R[1, 1])
-            y = math.atan2(-R[2, 0], sy)
-            z = 0
-
-        return np.array([x, y, z])
 
     def get_left_eye_width(self)->float:
         """Gets the left eye width
@@ -116,6 +130,10 @@ class Face():
         Returns:
             float: The width of the left eye
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         p1 = self.getlandmark_pos(self.left_eye_contour_indices[2])
         p2 = self.getlandmark_pos(self.left_eye_contour_indices[0])
         return np.abs(p2[0] - p1[0])
@@ -126,6 +144,10 @@ class Face():
         Returns:
             float: The height of the left eye
         """        
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         p1 = self.getlandmark_pos(self.left_eye_contour_indices[3])
         p2 = self.getlandmark_pos(self.left_eye_contour_indices[1])
         return np.abs(p2[1] - p1[1])
@@ -136,6 +158,10 @@ class Face():
         Returns:
             float: The width of the right eye
         """        
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         p1 = self.getlandmark_pos(self.right_eye_contour_indices[2])
         p2 = self.getlandmark_pos(self.right_eye_contour_indices[0])
         return np.abs(p2[0] - p1[0])
@@ -146,6 +172,10 @@ class Face():
         Returns:
             float: The height of the left eye
         """        
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         p1 = self.getlandmark_pos(self.right_eye_contour_indices[3])
         p2 = self.getlandmark_pos(self.right_eye_contour_indices[1])
         return np.abs(p2[1] - p1[1])
@@ -159,9 +189,14 @@ class Face():
         Returns:
             Tuple: Landmark 3D position in image space
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         lm = self.npLandmarks[index, ...]
         return (lm[0], lm[1], lm[2])
 
+
     def getlandmarks_pos(self, indices: list) -> np.ndarray:
         """Recovers the position of a landmark from a results array
 
@@ -171,8 +206,49 @@ class Face():
         Returns:
             np.ndarray: A nX3 array where n is the number of landmarks to be extracted and 3 are the 3 cartesian coordinates
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         return self.npLandmarks[indices,...]
 
+    def draw_landmark_by_index(self, image: np.ndarray, index: int, color: tuple = (255, 0, 0), radius: int = 5, thickness:int=1) -> np.ndarray:
+        """Draw a landmark on an image from landmark index
+
+        Args:
+            image (np.ndarray): Image to draw the landmark on
+            index (int): Index of the landmark
+            color (tuple, optional): Color of the landmark. Defaults to (255, 0, 0).
+            radius (int, optional): Radius of the circle to draw the landmark. Defaults to 5.
+            thickness (int, optional): Thickness of the line to draw the landmark. Defaults to 5.
+
+        Returns:
+            np.ndarray: Output image
+        """
+        pos = self.npLandmarks[index,:]
+        return cv2.circle(
+            image,(int(pos[0]), int(pos[1])), radius, color, thickness
+        )
+
+
+    def draw_landmarks(self, image: np.ndarray, landmarks: np.ndarray, radius:int=1, color: tuple = (255, 0, 0), thickness: int = 1) -> np.ndarray:
+        """Draw a list of landmarks on an image
+
+        Args:
+            image (np.ndarray): Image to draw the contour on
+            landmarks (np.ndarray): a nX3 ndarray containing the positions of the landmarks
+            radius (int, optional): Radius of the circle to draw the landmark. Defaults to 5.
+            color (tuple, optional): Color of the landmark. Defaults to (255, 0, 0).
+            thickness (int, optional): Thickness of the line to draw the landmark. Defaults to 5.
+
+
+        Returns:
+            np.ndarray: The image with the contour drawn on it
+        """
+        for i in range(landmarks.shape[0]):
+            image = cv2.circle(image, (int(landmarks[i,0]), int(landmarks[i,1])), radius,color, thickness)
+        return image
+
     def draw_landmark(self, image: np.ndarray, pos: tuple, color: tuple = (255, 0, 0), radius: int = 5, thickness:int=1) -> np.ndarray:
         """Draw a landmark on an image
 
@@ -184,11 +260,10 @@ class Face():
             thickness (int, optional): Thickness of the line to draw the landmark. Defaults to 5.
 
         Returns:
-            np.ndarray: [description]
+            np.ndarray: Output image
         """
         return cv2.circle(
-            image,
-            (int(pos[0]), int(pos[1])), radius, color, thickness
+            image,(int(pos[0]), int(pos[1])), radius, color, thickness
         )
 
     def draw_contour(self, image: np.ndarray, contour: np.ndarray, color: tuple = (255, 0, 0), thickness: int = 1) -> np.ndarray:
@@ -198,15 +273,14 @@ class Face():
             image (np.ndarray): Image to draw the contour on
             contour (np.ndarray): a nX3 ndarray containing the positions of the landmarks
             color (tuple, optional): Color of the landmark. Defaults to (255, 0, 0).
-            radius (int, optional): Radius of the circle to draw the landmark. Defaults to 5.
             thickness (int, optional): Thickness of the line to draw the landmark. Defaults to 5.
 
 
         Returns:
             np.ndarray: The image with the contour drawn on it
         """
-        pts = np.array([[int(p[0]), int(p[1])]
-                        for p in contour.tolist()]).reshape((-1, 1, 2))
+
+        pts = np.array([[int(p[0]), int(p[1])] for p in contour.tolist()]).reshape((-1, 1, 2))
         return cv2.polylines(image, [pts], True, color, thickness)
 
     def draw_overlay_on_left_iris(self, image:np.ndarray, overlay:np.ndarray)->np.ndarray:
@@ -219,6 +293,12 @@ class Face():
         Returns:
             np.ndarray: The image with the overlay drawn
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
+
+
         pImage = Image.fromarray(image)
         pos = self.getlandmark_pos(self.left_eye_center_index)[0:2]
 
@@ -242,6 +322,10 @@ class Face():
         Returns:
             np.ndarray: The image with the overlay drawn
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         pImage = Image.fromarray(image)
         pos = self.getlandmark_pos(self.right_eye_center_index)[0:2]
 
@@ -255,35 +339,59 @@ class Face():
             pImage.paste(overlay_, (x, y), overlay_)
         return np.array(pImage).astype(np.uint8)
 
-    def get_head_posture(self, orientation_style:int=0)->tuple:
+    def get_head_posture(self, camera_matrix:np.ndarray = None, dist_coeffs:np.ndarray=np.zeros((4,1)))->tuple:
         """Gets the posture of the head (position in cartesian space and Euler angles)
         Args:
-            orientation_style (int, optional) : Tells the style of orientation to be recovered:
-                                                0 : Rotation Matrix
-                                                1 : Euler angles in radians (Pitch, Yaw, Roll).
-                                                Defaults to 0 
+            camera_matrix (int, optional)       : The camera matrix built using buildCameraMatrix Helper function. Defaults to a perfect camera matrix 
+            dist_coeffs (np.ndarray, optional)) : The distortion coefficients of the camera
         Returns:
-            tuple: (position, orientation) the orientation is either in rotation matrix format (orientation_style==0) or in euler angles style : orientation_style==1
+            tuple: (position, orientation) the orientation is either in compact rodriguez format (angle * u where u is the rotation unit 3d vector representing the rotation axis). Feel free to use the helper functions to convert to angles or matrix
         """
-        # Head position
-        face_pos = self.npLandmarks.mean(axis=0)
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
+        if camera_matrix is None:
+            camera_matrix= buildCameraMatrix()
+        # V1 : 
+        # Method 1 Direct Linear transform (LDT)
+        # Now we decompose the facial cloud matrix multiplied by the reference facial cloud matrix to obtain the rotation matrix
+        """
+        lm = self.npLandmarks[Face.face_orientation_landmarks,:]
+        face_pos = lm.mean(axis=0)
 
         # Head orientation
-        facial_cloud = self.npLandmarks - face_pos
-
-        # If no reference was taken, then use this posture as the reference
-        if self.reference_facial_cloud is None:
-            self.reference_facial_cloud = facial_cloud
-
-        # Now we decompose the facial cloud matrix multiplied by the reference facial cloud matrix to obtain the rotation matrix
-        u, s, vh = np.linalg.svd(facial_cloud.T @ self.reference_facial_cloud)
+        facial_cloud = lm - face_pos
+        u, s, vh = np.linalg.svd(Face.face_3d_reference_positions.T @ face_2d_positions)
         R = vh @ u.T
-
         if orientation_style==0: # Rotation matrix
             face_ori = R
         elif orientation_style==1: # Euler angles in radians
             # Convert to euler angles
-            face_ori = self.rotationMatrixToEulerAngles(R)
+            face_ori = Face.rotationMatrixToEulerAngles(R)
+
+        """
+
+
+        # Head position
+        lm = self.npLandmarks[Face.face_orientation_landmarks,:]
+        face_pos = lm.mean(axis=0)
+
+
+        # V2 : Use opencv's PnPsolver to solve the rotation problem
+
+        face_2d_positions = self.npLandmarks[Face.face_orientation_landmarks,:2]
+        (success, face_ori, face_pos, _) = cv2.solvePnPRansac(
+                                                    Face.face_3d_reference_positions.astype(np.float),
+                                                    face_2d_positions.astype(np.float), 
+                                                    camera_matrix, 
+                                                    dist_coeffs,
+                                                    flags=cv2.SOLVEPNP_ITERATIVE)
+
+        if not success:
+            return None, None
+
+
 
         return face_pos, face_ori
 
@@ -293,6 +401,11 @@ class Face():
         Returns:
             int: The distance between the two eyes
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
+
         pos = self.getlandmarks_pos([self.left_eye_center_index, self.right_eye_center_index])
         return np.linalg.norm(pos[1,:]-pos[0,:])
 
@@ -313,35 +426,61 @@ class Face():
             left_eye_opening, right_eye_opening if blinking detection is deactivated
         """
 
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
+
         left_eye_center = self.getlandmark_pos(self.left_eye_center_index)
         left_eyelids_contour = self.getlandmarks_pos(self.left_eyelids_indices)
         left_eye_upper = left_eyelids_contour[3, ...]
         left_eye_lower = left_eyelids_contour[1, ...]
 
+        left_eye_contour = self.getlandmarks_pos(self.left_eye_contour_indices)
+        left_eye_iris_upper = left_eye_contour[3, ...]
+        left_eye_iris_lower = left_eye_contour[1, ...]
+
         right_eye_center = self.getlandmark_pos(self.right_eye_center_index)
         right_eyelids_contour = self.getlandmarks_pos(self.right_eyelids_indices)
         right_eye_upper = right_eyelids_contour[3, ...]
         right_eye_lower = right_eyelids_contour[1, ...]
 
+        right_eye_contour = self.getlandmarks_pos(self.right_eye_contour_indices)
+        right_eye_iris_upper = right_eye_contour[1, ...]
+        right_eye_iris_lower = right_eye_contour[3, ...]
+
+
+
+
         if draw_landmarks:
-            left_eye_contour = self.getlandmarks_pos(self.left_eye_contour_indices)
-            right_eye_contour = self.getlandmarks_pos(self.right_eye_contour_indices)
 
-            image = self.draw_contour(image, left_eyelids_contour, (0, 0, 0))
-            image = self.draw_contour(image, left_eye_contour, (0, 0, 0))
+            image = self.draw_landmark(image, left_eye_upper, (0, 0, 255),1)
+            image = self.draw_landmark(image, left_eye_lower, (0, 0, 255),1)
 
-            image = self.draw_landmark(image, left_eye_center, (255, 0, 255))
 
-            image = self.draw_contour(image, right_eyelids_contour, (0, 0, 0))
-            image = self.draw_contour(image, right_eye_contour, (0, 0, 0))
+            image = self.draw_landmark(image, left_eye_iris_upper, (255, 0, 0),1)
+            image = self.draw_landmark(image, left_eye_iris_lower, (255, 0, 0),1)
 
-            image = self.draw_landmark(image, right_eye_center, (255, 0, 255))
+            image = self.draw_landmark(image, right_eye_upper, (0, 0, 255),1)
+            image = self.draw_landmark(image, right_eye_lower, (0, 0, 255),1)
+
+            image = self.draw_landmark(image, right_eye_iris_upper, (255, 0, 0),1)
+            image = self.draw_landmark(image, right_eye_iris_lower, (255, 0, 0),1)            
+
+            image = self.draw_contour(image, left_eyelids_contour, (0, 0, 0),2)
+            image = self.draw_contour(image, left_eye_contour, (0, 0, 0),2)
+
+            image = self.draw_landmark(image, left_eye_center, (255, 0, 255),1)
+
+            image = self.draw_contour(image, right_eyelids_contour, (0, 0, 0),2)
+            image = self.draw_contour(image, right_eye_contour, (0, 0, 0),2)
+
+            image = self.draw_landmark(image, right_eye_center, (255, 0, 255),1)
 
 
 
         # Compute eye opening
-        left_eye_opening = np.linalg.norm(left_eye_upper-left_eye_lower) 
-        right_eye_opening = np.linalg.norm(right_eye_upper-right_eye_lower)
+        left_eye_opening = np.linalg.norm(left_eye_upper[0:2]-left_eye_lower[0:2])/np.linalg.norm(left_eye_iris_upper[0:2]-left_eye_iris_lower[0:2])
+        right_eye_opening = np.linalg.norm(right_eye_upper[0:2]-right_eye_lower[0:2])/np.linalg.norm(right_eye_iris_upper[0:2]-right_eye_iris_lower[0:2])
 
 
         if normalize:
@@ -377,6 +516,10 @@ class Face():
         Returns:
             list: A list of triangles extracted by Denaulay algorithm
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         if landmark_indices is None:
             landmarks = self.npLandmarks[:, :2]
         else:
@@ -423,6 +566,10 @@ class Face():
         Returns:
             np.ndarray: Image with triangles drawn on it
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
                 
         if landmark_indices is None:
             landmarks = self.npLandmarks[:, :2]
@@ -459,6 +606,10 @@ class Face():
         Returns:
             np.ndarray: Face drawn on a black background (the size of the image is equal of that of the face in the original image)
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         if landmark_indices is None:
             landmarks = self.npLandmarks[:, :2]
         else:
@@ -494,8 +645,12 @@ class Face():
             opacity (int, optional): the opacity level of the face (between 0 and 1)
 
         Returns:
-            np.ndarray: [description]
+            np.ndarray: An image containing only the face
         """
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
         # Get landmarks
         if landmark_indices is None:
             src_landmarks = self.npLandmarks[:, :2]
@@ -555,15 +710,37 @@ class Face():
         dst_image[int(dst_p1[1]):int(dst_p2[1]), int(dst_p1[0]):int(dst_p2[0])] = dst_crop
         return dst_image
 
-    def draw_mask(self, image:np.ndarray)->None:
+    def draw_bounding_box(self, image:np.ndarray, color:tuple=(255,0,0), thickness:int=1):
+        """Draws a bounding box around the face
+
+        Args:
+            image (np.ndarray): The image on which we will draw the bounding box
+            color (tuple, optional): The color of the bounding box. Defaults to (255,0,0).
+            thickness (int, optional): The line thickness. Defaults to 1.
+        """
+        pt1 = self.npLandmarks.min(axis=0)
+        pt2 = self.npLandmarks.max(axis=0)
+        cv2.rectangle(image, (int(pt1[0]),int(pt1[1])), (int(pt2[0]),int(pt2[1])), color, thickness)
+
+    def draw_mask(self, 
+                    image:np.ndarray, 
+                    landmarks_drawing_spec:DrawingSpec = DrawingSpec(color=(121, 0, 0), thickness=1, circle_radius=1), 
+                    contours_drawing_specs:DrawingSpec = DrawingSpec(color=(0, 0, 121), thickness=1, circle_radius=1),
+                    contour:frozenset=mp.solutions.face_mesh.FACEMESH_FACE_OVAL
+                    )->None:
         """Draws landmarks mask on a face
 
         Args:
             image (np.ndarray): Image to draw the mask on
         """
-        self.mp_drawing.draw_landmarks(image, self.landmarks, mp.solutions.face_mesh.FACEMESH_FACE_OVAL,
-                                       self.mp_drawing.DrawingSpec(
-                                           color=(80, 110, 10), thickness=1, circle_radius=1),
-                                       self.mp_drawing.DrawingSpec(
-                                           color=(80, 256, 121), thickness=1, circle_radius=1)
+
+        # Assertion to verify that the face object is ready
+        assert self.ready, "Face object is not ready. There are no landmarks extracted."
+
+
+        self.mp_drawing.draw_landmarks(image, self.landmarks , contour,
+                                       landmarks_drawing_spec,
+                                       contours_drawing_specs
                                        )
+
+
