@@ -12,5 +12,6 @@
             2 - process eyes and detect blinks, eye opening ...
             3 - do some fun face copying and pasting between images 
 <================"""
-from .Face import Face
-from .FaceAnalyzer import FaceAnalyzer
\ No newline at end of file
+from .Face import Face, DrawingSpec
+from .FaceAnalyzer import FaceAnalyzer
+from .Helpers import buildCameraMatrix, rodriguezToRotationMatrix
\ No newline at end of file
